(**************************************************************************

    This file is part of the LGSBloch package for Mathematica
    Copyright (C) 2009-2017
    Simon M. Rochester <simon@rochesterscientific.com>
    Ron Holzlöhner <rholzloe[at]eso.org>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)

+(* ::Section:: *)
(*Interface*)

BeginPackage["LGSBloch`LGSReturnFlux`", {"PhysicalConstants`", "Units`", "AtomicDensityMatrix`", "LGSBloch`"}];

Get["AtomicDensityMatrix`"];

setStates::usage = "";
makeAtomicSystem::usage = "";
makeExperimentalSystem::usage = "";
makeAtomicData::usage = "";
makeHamiltonian::usage = "";
makeLightHamiltonian::usage = "";
doRWA::usage = "";
makeRelaxationMatrix::usage = "";
makeSDampingMatrix::usage = "";
makeRepopulationMatrix::usage = "";
deleteHyperfineCoherences::usage = "";
makeVariableList::usage = "";
makeBlochEquations::usage = "";
makeBlochMatrix::usage = "";
makeFluorescenceOperator::usage = "";
findFluorescence::usage = "";
makeFluorescenceVector::usage = "";
makeTotalFluorescenceOperator::usage = "";
findTotalFluorescence::usage = "";
makeTotalFluorescenceVector::usage = "";
makeBlochBlocks::usage = "";
separateBlochBlocks::usage = "";

(* ::Section:: *)
(*Implementation*)


Begin["`Private`"];


(* ::Subsection:: *)
(*Setup*)


(* ::Text:: *)
(*Load the package, set options, and define utility functions*)


SetOptions[DensityMatrix, TimeDependence -> False, ComplexExpandVariables -> Subscript];


(* ::Text:: *)
(*Calculates the unperturbed resonance frequencies for an atomic system*)


ResonanceFrequencies[AtomicSystem_] := Module[{ha},
    ha = Hamiltonian[AtomicSystem, ElectricField -> {E0, 0, 0}];
    Simplify[
        Union@Flatten@MapIndexed[
            If[(Less @@ #2) && (#1 =!= 0), ha[[#2[[2]], #2[[2]]]] - ha[[#2[[1]], #2[[1]]]], {}]&,
            ha,
            {2}
        ] - Energy[2]
    ]
];


(* ::Text:: *)
(*Applies replacement rules to a sparse matrix*)


SparseReplace[expr_, reps : (_Rule | _RuleDelayed | _Dispatch | {___Rule | ___RuleDelayed})..] := Module[{AR},
    Fold[ReplaceAll, expr /. x_SparseArray :> AR[ArrayRules@x, Dimensions@x], {reps}] /. AR -> SparseArray
];


(* ::Text:: *)
(*Repeatedly performs replacement on itself*)


SelfReplaceRepeated[rules : {_Rule..}] := Thread[rules[[All, 1]] -> (rules[[All, 2]] //. rules)];


(* ::Subsection:: *)
(*Bloch equations setup*)


(* ::Text:: *)
(*Designation for the ground and excited states used in the AtomicData database.*)


setStates[atom_, line_] := {gstate, estate} = Switch[{atom, line},
    {"Na" | "Na23", "D1"},
    {{"Na", 23, {"Ne", {3, "s"}}, {2, "S", 1 / 2}}, {"Na", 23, {"Ne", {3, "p"}}, {2, "P", 1 / 2}}},
    {"Na" | "Na23", "D2"},
    {{"Na", 23, {"Ne", {3, "s"}}, {2, "S", 1 / 2}}, {"Na", 23, {"Ne", {3, "p"}}, {2, "P", 3 / 2}}},
    {"K" | "K39", "D1"},
    {{"K", 39, {"Ar", {4, "s"}}, {2, "S", 1 / 2}}, {"K", 39, {"Ar", {4, "p"}}, {2, "P", 1 / 2}}},
    {"K" | "K39", "D2"},
    {{"K", 39, {"Ar", {4, "s"}}, {2, "S", 1 / 2}}, {"K", 39, {"Ar", {4, "p"}}, {2, "P", 3 / 2}}},
    {"Rb87", "D1"},
    {{"Rb", 87, {"Kr", {5, "s"}}, {2, "S", 1 / 2}}, {"Rb", 87, {"Kr", {5, "p"}}, {2, "P", 1 / 2}}},
    {"Rb87", "D2"},
    {{"Rb", 87, {"Kr", {5, "s"}}, {2, "S", 1 / 2}}, {"Rb", 87, {"Kr", {5, "p"}}, {2, "P", 3 / 2}}},
    {"Rb85", "D1"},
    {{"Rb", 85, {"Kr", {5, "s"}}, {2, "S", 1 / 2}}, {"Rb", 85, {"Kr", {5, "p"}}, {2, "P", 1 / 2}}},
    {"Rb85", "D2"},
    {{"Rb", 85, {"Kr", {5, "s"}}, {2, "S", 1 / 2}}, {"Rb", 85, {"Kr", {5, "p"}}, {2, "P", 3 / 2}}},
    {"Cs" | "Cs133", "D1"},
    {{"Cs", 133, {"Xe", {6, "s"}}, {2, "S", 1 / 2}}, {"Cs", 133, {"Xe", {6, "p"}}, {2, "P", 1 / 2}}},
    {"Cs" | "Cs133", "D2"},
    {{"Cs", 133, {"Xe", {6, "s"}}, {2, "S", 1 / 2}}, {"Cs", 133, {"Xe", {6, "p"}}, {2, "P", 3 / 2}}},
    _,
    Print["Invalid atom or transition."];
    Abort[];
];


(* ::Text:: *)
(*Generate a list of the Zeeman sublevels of the system.*)


makeAtomicSystem[] := atsys = Sublevels@{
    AtomicState[1, Join[AtomicData[gstate, {J, L, S, Energy, NaturalWidth, NuclearSpin}], {}]],
    AtomicState[2, Join[AtomicData[estate, {J, L, S, NuclearSpin}], {BranchingRatio[1] -> 1}]]
};


(* ::Text:: *)
(*Make a list of numerical parameters describing the system. All frequencies, widths, and rates are in omega units. *)
(*RMSVelocity is rms atomic velocity in meter/sec*)
(*LI is light intensity in watt/meter^2*)
(*T is temperature in Kelvin*)
(*LandeGF is g-factor of highest-angular-momentum ground state (first state in the atsys list).*)

makeAtomicData[] := (
    constants = {c -> SpeedOfLight, \[HBar] -> PlanckConstantReduced, \[Epsilon] -> VacuumPermittivity};
    (* reduced dipole matrix element in ADM units *)
    dADM = Convert[ExpandDipoleRME[atsys, ReducedME[1, {Dipole, 1}, 2]] /. AtomicData[estate, StateLabel -> 2], Second];
    (* convert reduced dipole matrix element to cgs by multiplying by Sqrt[hbar c^3], and then to SI
       by multiplying by Sqrt[4 pi epsilon] *)
    dSI = Convert[dADM Sqrt[\[HBar] c^3] Sqrt[4 \[Pi] \[Epsilon]] /. constants, Coulomb Meter];
    (* electric field amplitude in SI in terms of light intensity in W/m^2 *)
    ESI[i_] = Convert[Sqrt[2. i (Watt/Meter^2)/(c \[Epsilon])] /. constants, Volt/Meter];
    (* Rabi frequency in terms of light intensity *)
    rabi[i_] = Convert[ESI[i] dSI/\[HBar] /. constants, 1/Second];
    atdata = Flatten@{
        {
            AtomicData[gstate, {HyperfineA}, StateLabel -> 1],
            AtomicData[estate, {HyperfineA, HyperfineB, NaturalWidth, Energy}, StateLabel -> 2]
        },
        {
            DopplerWidth -> Convert[DopplerWidth[gstate, estate, Temperature], 1 / Second],
            RMSVelocity -> Convert[MostProbableSpeed[estate, Temperature] / Sqrt[2], Meter / Second],
            RabiFrequencyA[b_] -> rabi@LightIntensityA[b],
            RabiFrequencyB[b_] -> rabi@LightIntensityB[b]
        },
        BohrMagneton -> Convert[ElectronCharge / (2 ElectronMass), Mega / Second / Gauss],
        LandeGF -> LandeGF@First@atsys
    } /. {Mega -> 10^6, Second -> 1, Meter -> 1, Second -> 1, Watt -> 1};
    AppendTo[atdata, ResonanceFrequencies -> Sort[ResonanceFrequencies@atsys /. atdata]]
);


(* ::Text:: *)
(* Define the environment that the atoms are subject to. A ExperimentalRegion is a designator for a group of atoms all
   under the same conditions. In this case, the Regions are n velocity groups of the Doppler distribution, with Doppler
   shift DopplerShift[i] and relative population VGDensity[i] . We could set up any number of velocity groups here, but
   that would involve a lot of redundant computation, since the equations for each group are similar. Instead, we will
   set up two velocity groups here, and use the resulting equations as building blocks to put together the evolution
   equations for the entire system. *)


(* ::Text:: *)
(*\[Bullet] TransitRate[VoidRegion] is the rate at which atoms enter and leave the light beam.*)
(*\[Bullet] TransitRate[r1] is the rate at which atoms are transferred from the particular velocity group r to another velocity group r1. This depends on the overall collision rate VccRate, and the fraction of atoms VGDensity[r1] in the velocity group r1.*)
(*\[Bullet] Density is the atomic density for atoms in a given velocity group. AtomicDensity is the overall atomic density, and VGDensity[r] is the fraction in that group.*)
(*\[Bullet] MagneticField is the magnetic field vector in the Cartesian basis. It is specified in terms of the Larmor frequency \[CapitalOmega]L for the F=2 ground state, and its spherical angles \[Theta], \[Phi].*)
(*\[Bullet] Optical parameters are the parameters for the laser beam: {frequency, electric field, {polarization angle, ellipticity}}. *)
(*	The propagation  and reference polarization directions can be specified through the options PolarizationVector and PropagationVector. *)
(*	If they are not, the default values PolarizationVector -> {1, 0, 0}, PropagationVector -> {0, 0, 1} are used.*)
(*	An ellipticity angle value of zero corresponds to linear polarization, \[PlusMinus]\[Pi]/4 corresponds to circular polarization. A symbolic value can be used here and a numerical value substituted when a calculation is performed.	*)


(* ::Text:: *)
(*WHAT HAPPENS TO OPTICAL COHERENCES DURING VELOCITY-CHANGING COLLISIONS OR WHEN TRAVELING BETWEEN DITHERING REGIONS?*)


makeExperimentalSystem[pulsedSetup_, flipDoppler_] := expsys = Flatten@Table[
    ExperimentalRegion[{r, b}, {
        TransitRate[VoidRegion] -> BeamTransitRate,
        TransitRate[{rother_, bother_}] -> Which[
            bother == b,
            Evaluate[VccRate[r, rother]],
            Evaluate[rother == r],
            DitherRegionFraction[bother] DitherRate,
            True,
            0
        ],
        Density -> AtomicDensity VGDensity[r] DitherRegionFraction[b],
        OpticalParameters -> {
            Energy[2] + \[CapitalDelta][b] - If[flipDoppler, -1, 1] DopplerShift[r] + If[pulsedSetup, ModeDetuning, 0],
            {\[CapitalOmega]R[b] / ReducedME[1, {Dipole, 1}, 2], LightPhase[b]},
            {0, \[Epsilon]}
        },
        MagneticField -> LarmorFrequency {
            Cos[MagneticAzimuth] Sin[MagneticZenith],
            Sin[MagneticAzimuth] Sin[MagneticZenith],
            Cos[MagneticZenith]
        } / ((LandeGF /. atdata) BohrMagneton)
    }],
    {b, 2},
    {r, 2}
];


(* ::Text:: *)
(*The Breit-Rabi Hamiltonian. *)
(*DOES NOT USE EXPSYS, SO WILL NEED TO BE FIXED IF MAGNETIC FIELDS IN EXPSYS ARE CHANGED.*)


BreitRabi[s_AtomicState, B_] /; J[s] === 1 / 2 && NuclearSpin[s] > 0 :=
    -(\[CapitalDelta] / (2(2N + 1))) + j \[CapitalDelta] / 2 (1 + (4m \[Xi]) / (2N + 1) + \[Xi]^2)^(1 / 2) /.
        \[Xi] -> LandeGJ[s] BohrMagneton B / \[CapitalDelta] /.
        \[CapitalDelta] -> 2 HyperfineA[s] /.
        N -> NuclearSpin[s] /.
        m -> M[s] /.
        j -> If[F[s] == NuclearSpin[s] + 1 / 2, 1, -1];


BRHamiltonian[] := WignerRotate[
    atsys,
    DiagonalMatrix[If[GroundStateQ@#, BreitRabi[#, LarmorFrequency / ((LandeGF /. atdata) BohrMagneton)], 0]& /@ atsys],
    {MagneticAzimuth, MagneticZenith, 0},
    StandardOperator -> True
];


(* ::Text:: *)
(*Generate the Hamiltonians for the atoms in the two velocity groups, subject to the magnetic field.*)


makeHamiltonian[neglectExcitedStateMixing_, useBreitRabi_] := (
    H = Hamiltonian[atsys, expsys, Interaction -> {Internal, Magnetic}];
    BRH = If[useBreitRabi, BRHamiltonian[], 0];
    H1 = (BRH + # (
        DensityMatrix[atsys, ComplexExpandVariables -> False] /.
            If[useBreitRabi,
                DMElementPattern[ComplexExpandVariables -> False] /; StateLabel1 == StateLabel2 == 1 -> 0,
                {}
            ] /.
            DMElementPattern[ComplexExpandVariables -> False] /; StateLabel1 == StateLabel2 == 1 && F1 != F2 -> 0 /.
            If[neglectExcitedStateMixing,
                DMElementPattern[ComplexExpandVariables -> False] /; StateLabel1 == StateLabel2 == 2 && F1 != F2 -> 0,
                {}
            ] /. DMElementPattern[ComplexExpandVariables -> False] -> 1))& /@ H
);


(* ::Text:: *)
(*Generate the Hamiltonians for the light.*)


(* ::Text:: *)
(* Split up the light field Hamiltonian into two fields, one of which only interacts with the D2a (ground F=I+1/2)
   transition, and the other only interacts with the D2b (ground F=I-1/2) transition, in order to model the repump
   light. *)
(* THIS APPROXIMATION IS NOT VALID IF THE LIGHT FREQUENCIES STRAY TOO FAR FROM RESONANCE. *)


makeLightHamiltonian[] := (
    Hlight = Hamiltonian[atsys, expsys, Interaction -> {E1}];
    Hlighta = DirectSelfProduct[If[StateLabel[#] == 1 && F[#] == NuclearSpin[#] - J[#], 0, 1]& /@ atsys]#& /@ Hlight /.
        {\[Epsilon] -> EllipticityA, \[CapitalDelta] -> DetuningA, \[CapitalOmega]R -> RabiFrequencyA};
    Hlightb = DirectSelfProduct[If[StateLabel[#] == 1 && F[#] == NuclearSpin[#] + J[#], 0, 1]& /@ atsys]#& /@ Hlight /.
        {\[Epsilon] -> EllipticityB, \[CapitalDelta] -> DetuningB, \[CapitalOmega]R -> RabiFrequencyB};
);


(* ::Text:: *)
(*Perform the rotating wave approximation on the Hamiltonians.*)


doRWA[flipDoppler_] := (
    Hrwa = Table[
        {vgn, beamn} = RegionLabel[expsys[[i]]];
        sign = If[OrderedQ[{"09.12.08", $ADMVersion}], -1, 1];
        tm = MatrixExp[
            sign I t DiagonalMatrix[
                Which[
                    StateLabel[#] == 1 && F[#] == NuclearSpin[#] + J[#],
                    -(DetuningA[beamn] + Energy[2] - If[flipDoppler, -1, 1]DopplerShift[vgn]),
                    StateLabel[#] == 1 && F[#] == NuclearSpin[#] - J[#],
                    -(DetuningB[beamn] + Energy[2] - If[flipDoppler, -1, 1]DopplerShift[vgn]),
                    True,
                    0
                ]& /@ atsys]
        ];
        RotatingWaveApproximation[atsys, (H1 + Hlighta + Hlightb)[[i]], Energy[2], TransformMatrix -> tm],
        {i, 4}
    ]
);


(* ::Text:: *)
(* Generate the matrices describing relaxation due to spontaneous decay, and effective relaxation due to atom transit
   of the light beam and velocity changing collisions, which cause atoms to leave one velocity group and be transferred
   to another. *)


makeRelaxationMatrix[] := (rel = Relaxation[atsys, expsys]);


(* ::Text:: *)
(*Generate the matrix describing S-damping of the ground state: J^2 \[Rho]-J\[CenterDot](\[Rho] J). *)
(*WignerEckart[atsys, {J, 1}] generates the matrix representation of the spherical components of the J operator for the specified atomic system.*)


makeSDampingMatrix[] := (
    gs = WignerEckart[atsys, {GroundState, 0, 0}];
    jcart = ToCartesian@WignerEckart[atsys, {J, 1}];
    dm = DensityMatrix[atsys, DMLabel -> temp];
    jsqr = Simplify[jcart\[CenterDot]jcart];
    sd = SDampingRate gs.(jsqr.dm - jcart\[CenterDot]((dm.#)& /@ jcart)).gs // Expand;
    sdamp = (sd /. temp -> RegionLabel[#])& /@ expsys
);


(* ::Text:: *)
(* Generate the matrices describing the repopulation of the ground state due to spontaneous decay, and the appearance
   of atoms in each velocity group due to atoms entering the light beam and changing their velocity from another
   velocity group. *)


(* ::Text:: *)
(*SHOULD DECIDE WHAT HAPPENS TO EXCITED-STATE ATOMS AND OPTICAL COHERENCES DURING VELOCITY-CHANGING COLLISIONS.*)


makeRepopulationMatrix[] := (
    rep = Expand@Repopulation[atsys, expsys] /.
        x : (DitherRate DMElementPattern[]) /; StateLabel1 != StateLabel2 -> DitherCoherenceTransfer x;
    recoil = rep /. {BeamTransitRate -> 0, VccRate[__] -> 0, DitherRate -> 0} /.
        Subscript[\[Rho], re_, {1, b_}, s1_, s2_] -> 0 /.
        Subscript[\[Rho], re_, {v_, b_}, s1_, s2_] -> Subscript[\[Rho], re, {v - 1, b}, s1, s2] /.
        NaturalWidth[2] -> RecoilShift / dv NaturalWidth[2];
    rep1 = MapThread[
        #1 /.
            If[RegionLabel[#2][[1]] == 1,
                {NaturalWidth[2] -> (1 - RecoilShift / dv) NaturalWidth[2]},
                {}
            ]&,
        {rep, expsys}
    ] - (LaserWidth DensityMatrix[atsys, expsys] / 2 /. DMElementPattern[] /; StateLabel1 == StateLabel2 -> 0)
);


(* ::Text:: *)
(*Use the Hamiltonian, relaxation, and repopulation matrices in the Liouville equation to produce the Bloch equations.*)


(* ::Text:: *)
(*Here we set ground-state and excited-state hyperfine coherences to zero. SHOULD QUANTIFY WHEN THIS APPROXIMATION IS VALID.*)


deleteHyperfineCoherences[] := delpos = Position[
    DMVariables[atsys, expsys],
    DMElementPattern[] /; (StateLabel1 == StateLabel2 && F1 != F2) || (StateLabel1 != StateLabel2 && Abs[F1 - F2] > 1),
    1
];


(* ::Text:: *)
(*A list of the variables describing the density matrix elements.*)


makeVariableList[] := {delvars = Extract[DMVariables[atsys, expsys], delpos], vars = Delete[DMVariables[atsys, expsys], delpos]};


makeBlochEquations[] := BlochEqs = Delete[
    LiouvilleEquation[atsys, expsys, Hrwa, rel, rep1 + recoil - sdamp],
    delpos
] /. Dispatch[((# -> 0)& /@ delvars)];


(* ::Text:: *)
(* Produce the equivalent matrix describing the Bloch equations. I refer to this as the "Bloch matrix". This can also
   be thought of as the matrix for the evolution operator in the "Liouville" vector representation of the density matrix.*)


makeBlochMatrix[] := BlochMat = Chop@Reverse[{-1, 1} CoefficientArrays[BlochEqs[[All, 2]], vars]];


(* ::Text:: *)
(*Produce the operator describing fluorescence into the solid angle at \[Theta]=0, \[Phi]=0. The Fluorescence operator
  is in units of power/sr, so we divide by photon energy to get units of photons/s/sr.*)


makeFluorescenceOperator[] := flop = Simplify@FluorescenceOperator[atsys, Energy[2], ViewTheta, ViewPhi] / Energy[2];


(* ::Text:: *)
(*Find the expectation value of the fluorescence operator to get an expression for the fluorescence in terms of the density matrix elements.*)


findFluorescence[] := fluor = Total[Tr[flop.#]& /@ DensityMatrix[atsys, expsys]];


(* ::Text:: *)
(*Convert the expression for fluorescence into a vector suitable for dotting with the density matrix in vector form (aka in "Liouville space").*)


makeFluorescenceVector[] := fluorvec = Normal@CoefficientArrays[fluor, vars][[2]];


makeTotalFluorescenceOperator[] := totflop = Simplify@FluorescenceOperator[atsys, Energy[2]] / Energy[2];


findTotalFluorescence[] := totfluor = Total[Tr[totflop.#]& /@ DensityMatrix[atsys, expsys]];


makeTotalFluorescenceVector[] := totfluorvec = Normal@CoefficientArrays[totfluor, vars][[2]];


(* ::Text:: *)
(* We now chop up the Bloch matrix into submatrices for each velocity group and subbeam (or, for the off diagonal
   blocks, the interaction between two velocity groups or beams.) We will then put these blocks together to create a
   matrix describing as many velocity groups and beams as required. *)


makeBlochBlocks[] := (
    nvars = Length[vars] / Length[expsys];
    vars1 = Take[vars, nvars];
    A11 = SparseReplace[
        BlochMat[[1, ;; nvars, ;; nvars]],
        {VGDensity[2] -> (1 - VGDensity[1]), DitherRegionFraction[2] -> 1 - DitherRegionFraction[1]}
    ];
    A12vg = BlochMat[[1, ;; nvars, nvars + 1 ;; 2 nvars]];
    A21vg = BlochMat[[1, nvars + 1 ;; 2 nvars, ;; nvars]];
    A12beam = BlochMat[[1, ;; nvars, 2 nvars + 1 ;; 3 nvars]];
    A21beam = BlochMat[[1, 2 nvars + 1 ;; 3nvars, ;; nvars]];
    b1 = BlochMat[[2, 1 ;; nvars]] // Normal;
    fluorvec1 = fluorvec[[;; nvars]] // Normal;
    totfluorvec1 = totfluorvec[[;; nvars]] // Normal;
);


(* ::Text:: *)
(* The only difference between the blocks for each velocity group is the values of the Doppler shift DopplerShift[i]
   and the relative density VGDensity[i]. We create separate matrices describing the dependence on each of these
   parameters, allowing more efficient manipulation of the submatrices. *)


separateBlochBlocks[] := (
    A110 = SparseReplace[A11, {DopplerShift[1] -> 0, VccRate[1, 2] -> 0, RecoilShift -> 0}];
    A11S = SparseArray @@ (
        Expand[{ArrayRules[A11 - SparseReplace[A11, {DopplerShift[1] -> 0}]], Dimensions[A11]}] /.
            ({x_Integer, y_Integer} -> 0) -> Sequence[]
    ) / DopplerShift[1];
    A11G = SparseArray @@ (
        Expand[{ArrayRules[A11 - SparseReplace[A11, {VccRate[1, 2] -> 0}]], Dimensions[A11]}] /.
            ({x_Integer, y_Integer} -> 0) -> Sequence[]
    ) / VccRate[1, 2];
    A11R = SparseArray @@ (
        Expand[{ArrayRules[A11 - SparseReplace[A11, {RecoilShift -> 0}]], Dimensions[A11]}] /.
            ({x_Integer, y_Integer} -> 0) -> Sequence[]
    ) dv;
    A12G = A12vg / ((VccRate[1, 2] VGDensity[1]) / VGDensity[2]);
    A21G = SparseReplace[A21vg, RecoilShift -> 0] / ((VccRate[2, 1] VGDensity[2]) / VGDensity[1]);
    A21R = SparseReplace[A21vg, VccRate[2, 1] -> 0]dv;
    A12B = A12beam / DitherRegionFraction[1];
    A21B = A21beam / DitherRegionFraction[2];
    b1G = b1 / VGDensity[1];
);


Options[GenerateLGSSystem] = {NeglectHyperfineCoherences -> True, PrintDiagnostics -> False,
    NeglectExcitedStateMixing -> True, PulsedSetup -> False, FlipDoppler -> False};


GenerateLGSSystem[atom_String : "Na", line_String : "D2", OptionsPattern[]] := Module[{diag},
    diag = OptionValue@PrintDiagnostics;
    SetOptions[DensityMatrix, TimeDependence -> False, ComplexExpandVariables -> Subscript];
    setStates[atom, line];
    makeAtomicSystem[];
    If[diag, Print@atsys];
    makeAtomicData[];
    If[diag, Print@atdata];
    makeExperimentalSystem[OptionValue@PulsedSetup, OptionValue@FlipDoppler];
    If[diag, Print@expsys];
    makeHamiltonian[OptionValue@NeglectExcitedStateMixing, False];
    If[diag, Print[MatrixForm /@ H1]];
    makeLightHamiltonian[];
    doRWA[OptionValue@FlipDoppler];
    If[diag, Print[MatrixForm /@ Hrwa]];
    makeRelaxationMatrix[];
    makeSDampingMatrix[];
    makeRepopulationMatrix[];
    If[OptionValue@NeglectHyperfineCoherences, deleteHyperfineCoherences[], delpos = {}];
    makeVariableList[];
    makeBlochEquations[];
    makeBlochMatrix[];
    If[diag, Print@MatrixPlot[First@BlochMat, Mesh -> {3, 3}, MaxPlotPoints -> 2400]];
    makeFluorescenceOperator[];
    findFluorescence[];
    makeFluorescenceVector[];
    makeTotalFluorescenceOperator[];
    findTotalFluorescence[];
    makeTotalFluorescenceVector[];
    makeBlochBlocks[];
    separateBlochBlocks[];
    {nvars, A110, A11S, A11G, A11R, A12G, A21R, A12B, b1G, fluorvec1,
        Sequence @@ If[OptionValue[PulsedSetup], {totfluorvec1, vars, BlochEqs}, {}], atdata}
];


(* ::Subsection:: *)
(*VelocityGroups1*)


(* ::Text:: *)
(* Here we define the function VelocityGroups, which finds an appropriate set of atomic velocity bins for performing
   the calculation. It creates a set of narrow bins near the resonance frequencies, and wider bins to cover the rest of
   the Doppler distribution.*)


(* ::Text:: *)
(*It is called with*)
(* VelocityGroups[DopplerWidth, ResonanceFrequencies, {SmallBinSize, n1}, {LargeBinSize, n2}]*)
(*where*)
(* DopplerWidth is the Doppler width*)
(* ResonanceFrequencies are the optical resonance frequencies relative to the light frequency (i.e. transition frequencies in the rotating frame)*)
(* SmallBinSize is the velocity group bin size for bins near the resonance frequencies*)
(* n1 specifies that (2 n1+1) bins are created near each resonance frequency*)
(* LargeBinSize is the maximum size for a bin not near a resonance frequency*)
(* n2 is the multiple of the Doppler width that determines the range over which velocity groups should be created*)
(* If ShowPlot->True is specified, a plot is produced that shows the limits of the velocity group bins (blue lines),
   and the weighted centers and number of atoms in each bin (black lines). *)


(* ::Text:: *)
(*We can probably do better by using an adaptive procedure that adjusts the bin sizes in response to the calculated signal due to each velocity group.*)


Off[Solve::"ifun"];


intG[\[CapitalGamma]D_, a_, b_] = Integrate[Exp[-(\[CapitalDelta] / \[CapitalGamma]D)^2] / (Sqrt[\[Pi]] \[CapitalGamma]D), {\[CapitalDelta], a, b}, Assumptions -> {\[CapitalGamma]D > 0, b > a > 0}];


(*CDFGaussian[\[CapitalGamma]D_, b_] = Integrate[Exp[-(\[CapitalDelta] / \[CapitalGamma]D)^2] / (Sqrt[\[Pi]] \[CapitalGamma]D), {\[CapitalDelta], -Infinity, b}, Assumptions -> {\[CapitalGamma]D > 0, b\[Element]Reals}];*)


gaussian[\[CapitalGamma]_, \[CapitalDelta]_] = Exp[-(\[CapitalDelta] / \[CapitalGamma])^2] / (Sqrt[\[Pi]] \[CapitalGamma]);


inverseIntG[\[CapitalGamma]D_, a_, n_] = b /. Solve[n == intG[\[CapitalGamma]D, a, b], b][[1]];


binCenter[\[CapitalGamma]D_, {a_, b_}] := {inverseIntG[\[CapitalGamma]D, a, # / 2], #}&@intG[\[CapitalGamma]D, a, b];


Options[VelocityGroups1] = {ShowPlot -> True};


fitRange[rmin_, rmax_, bin_] := Range[rmin, rmax, (rmax - rmin) / Max[Round[Abs[(rmax - rmin) / bin]], 1]];


VelocityGroups1[GD_, res_, {binsize1_, n1_}, {binsize2_, n2_}, OptionsPattern[]] := Module[{res1, bins0, bins1, bins2, bins, groups},
    res1 = Sort@Union@Flatten@{-n2 GD, res, n2 GD};
    res1 = Mean /@ Split[res1, #2 - #1 < n1 binsize1&];
    bins0 = Select[
        Sort@Flatten@Table[{
            Range[
                res1[[i]] - .5 binsize1,
                Max[res1[[i]] - (n1 + .5) binsize1, Mean@{res1[[i]], res1[[i - 1]]}, -n2 GD],
                -binsize1
            ],
            Range[
                res1[[i]] + .5 binsize1,
                Min[res1[[i]] + (n1 + .5) binsize1, Mean@{res1[[i]], res1[[i + 1]]}, n2 GD],
                binsize1
            ]},
        {i, 2, Length[res1] - 1}
        ],
        -n2 GD <= # <= n2 GD&
    ];
    bins1 = Union@Flatten@{-n2 GD, bins0, n2 GD};
    bins1 = Mean /@ Split[bins1, #2 - #1 < binsize1&];
    bins2 = Union@Flatten@{
        Table[Most@fitRange[bins1[[i]], bins1[[i + 1]], binsize2], {i, Length[bins1] - 1}],
        Last@bins1
    };
    bins = Transpose@{Prepend[bins2, -Infinity], Append[bins2, Infinity]};
    groups = binCenter[GD, #]& /@ bins;
    If[OptionValue[ShowPlot],
        Print@Show[Plot[GD gaussian[GD, \[CapitalDelta]], {\[CapitalDelta], -3GD, 3GD}, PlotRange -> All],
            ListPlot[groups, Filling -> 0, FillingStyle -> Directive[Black], Axes -> False, PlotRange -> All],
            ListPlot[{#, GD gaussian[GD, #]}& /@ Flatten[bins], Filling -> 0, PlotStyle -> None]
        ]];
    MapThread[Append[#1, #2]&, {groups, bins}]
];


(* ::Subsection:: *)
(*VelocityGroups2*)


(* ::Text:: *)
(*A second method for choosing velocity groups, using an estimate for the spectrum to make a weighting function.*)


lorentzian[a_, x_] := a / (2\[Pi]) 1 / ((a / 2)^2 + x^2);


gaussian[a_, x_] := 1 / (a Sqrt[\[Pi]]) Exp[-(x / a)^2];


nNIntegrate[f_, {x_, x0_, x1_}] := Module[{int},
    int = y[x] /. First@NDSolve[{y'[x] == f, y[x0] == 0}, y[x], {x, x0, x1}, MaxStepFraction -> 1 / 50000, MaxSteps -> 10^6];
    Piecewise[{{int / (int /. x -> x1), x0 <= x <= x1}, {1, x > x1}}]
];


WeightFunction[sig_, {a_, b_, c_}, {x_, x0_, x1_}] :=
    a (x - x0) / (x1 - x0) + b nNIntegrate[sig, {x, x0, x1}] + c nNIntegrate[Abs[D[sig, {x, 2}]], {x, x0, x1}];


Options[VelocityGroups2] = {ShowPlot -> True};


VelocityGroups2[wint_, {x_, first_, last_}, n_, dw_, minstep_, freqs_, OptionsPattern[]] := Module[
    {bins, groups, points, edges, pops, wint1},
    wint1[x1_] = (wint /. x -> x1);
    points = NestWhileList[
        Min[last, Max[# + minstep, (x1 /. FindRoot[wint1[x1] - wint1[#] - wint1[last] / n, {x1, #, last}])]]&,
        first,
        # < last - (last - first)10^-6&
    ];
    edges = Join[{-Infinity}, Mean /@ Transpose[{Most@points, Rest@points}], {Infinity}];
    bins = Transpose[{Most@edges, Rest@edges}];
    pops = intG[dw, #[[1]], #[[2]]]& /@ bins;
    groups = Transpose[{points, pops, bins}];
    If[OptionValue[ShowPlot],
        Print@Show[
            Plot[dw gaussian[dw, \[CapitalDelta]], {\[CapitalDelta], first, last}, PlotRange -> All, Frame -> True, Axes -> False],
            ListPlot[groups[[All, 1 ;; 2]], Axes -> False, PlotRange -> All],
            ListLinePlot[
                Join[
                    {{groups[[1, 3, 1]], 0}},
                    Flatten[{{#[[3, 1]], #[[2]]}, {#[[3, 2]], #[[2]]}}& /@ groups, 1],
                    {{groups[[-1, 3, 2]], 0}}
                ],
                Filling -> 0
            ],
            ListPlot[{#, dw gaussian[dw, #]}& /@ Flatten[groups[[All, -1]]], Filling -> 0, PlotStyle -> None]
        ]
    ];
    groups
];


bend[{p0 : {x0_, y0_}, p1 : {x1_, y1_}, p2 : {x2_, y2_}}] := (Slope[{p1, p2}] - Slope[{p0, p1}]) / ((x2 - x0) / 2);


bendList[list_] := Module[{bl, max},
    bl = Join[
        {{list[[1, 1]], 0}},
        Table[{list[[i + 1, 1]], bend[list[[i ;; i + 2]]]}, {i, Length[list] - 2}],
        {{list[[-1, 1]], 0}}
    ];
    max = Max[Abs@bl[[All, 2]]];
    {1, 1 / max}#& /@ bl
];


nWeightFunction[spec_, bend_, {a_, b_, c_}, {x_, x0_, x1_}] := Module[
    {specint = Integrate[spec[x], x],
        absbend = FunctionInterpolation[Abs[bend[x]], {x, bend[[1, 1, 1]], bend[[1, 1, 2]]}, InterpolationOrder -> 1],
        bendint},
    bendint = Integrate[absbend[x], x];
    a (x - x0) / (x1 - x0) +
        b Piecewise[{{0, x < x0}, {specint / (specint /. x -> x1), x0 <= x <= x1}, {1, x > x1}}] +
            c Piecewise[{{0, x < x0}, {bendint / (bendint /. x -> x1), x0 <= x <= x1}, {1, x > x1}}]
];


fakeSignalGroups[] := (
    sig[x_] = NaturalWidth[2] DopplerWidth *
        Total[(lorentzian[NaturalWidth[2] + \[CapitalOmega]R^2 / NaturalWidth[2], x - (# - \[CapitalDelta])]& /@ Drop[freqs, 2])] *
            gaussian[DopplerWidth, x] /. params;
    wf = WeightFunction[sig@x, {1, 1, 1}, {x, -3dw, 3dw}];
    vg = VelocityGroups2[wf, {x, -3dw, 3dw}, nog[[i]], dw, ShowPlot -> False];
);


(* ::Subsection:: *)
(*Solve functions*)


(* ::Text:: *)
(* We first find an appropriate set of velocity groups using the VelocityGroups function: *)
(* VelocityGroups[DopplerWidth, ResonanceFrequencies, {SmallBinSize, n1}, {LargeBinSize, n2}] *)
(* where *)
(* ResonanceFrequencies are the optical resonance frequencies relative to the light frequency *)
(* SmallBinSize is the velocity group bin size for bins near the resonance frequencies *)
(* n1 specifies that (2 n1+1) bins are created near each resonance frequency *)
(* LargeBinSize is the minimum size for a bin not near a resonance frequency *)
(* n2 is the multiple of the doppler width that determines the range over which velocity groups should be created *)
(* If ShowPlot->True is specified, a plot is produced that shows the limits of the velocity group bins (blue lines),
   and the weighted centers and number of atoms in each bin (black lines). *)


(* ::Text:: *)
(* The parameters used in VelocityGroups here are sufficient to obtain reasonable results, but more velocity groups are
  needed for complete accuracy. *)


(* ::Text:: *)
(* We then substitute the specified numerical parameters into the submatrix describing one velocity group, and create
   a large matrix using copies of the submatrices. This matrix is solved using LinearSolve, and the solution is dotted
   with the vector describing fluorescence. The Krylov method, an iterative method, is used in LinearSolve. A
   preconditioner consisting of the diagonal blocks of the Bloch matrix is used. *)


Options[ReturnFlux] = {RefineResult -> {}, InitVGParams -> {{1, 2}, {1, 3}}, FluxReport -> {}, ReturnAllSteps -> False,
    PlotVelocityGroups -> False, InitVG -> Automatic, SkipLinearSolve -> False};


massageParams[par_, atdata_] := Module[{nb = NDither /. par},
    Dispatch@SelfReplaceRepeated@Flatten@{
        atdata,
        Table[
            MapThread[Rule, {
                {DitherRegionFraction[i], DetuningA[i], LightIntensityA[i], DetuningB[i], LightIntensityB[i]},
                {1 / nb,
                    DetuningA + Sign[i - (nb + 1) / 2] DitherDetuning[Abs[i - (nb + 1) / 2]] / 2,
                    LightIntensityA,
                    DetuningB + Sign[i - (nb + 1) / 2] DitherDetuning[Abs[i - (nb + 1) / 2]] / 2,
                    LightIntensityB
                } /. {DitherDetuning[n_ /; n \[NotElement] Integers] -> DitherDetuning[n + 1 / 2]} /. par}],
            {i, nb}
        ],
        DeleteCases[par, (DetuningA -> _) | (DetuningB -> _) | (LightIntensityA -> _) | (LightIntensityB -> _)]}
];


ReturnFlux[sys_, par_, OptionsPattern[]] := Module[{nb, params, dw, nog, allfreqs, vg, atdata, freqs, speclist, nvars,
    resultTable, spec, bendfunc, wf, solve, rep, defaultViewDirection, fluxReport = OptionValue@FluxReport},
    nb = NDither /. par;
    atdata = Last@sys;
    nvars = First@sys;
    params = massageParams[par, atdata];
    dw = DopplerWidth /. params;
    nog = OptionValue@RefineResult /. False | None -> {};
    rep = EnsureList[fluxReport /. False | None -> {}];
    freqs = ResonanceFrequencies /. atdata;
	defaultViewDirection = Flatten@MapThread[
        If[!NumericQ[#1 /. params], (#1 /. params) -> #2, {}]&,
        {{ViewTheta, ViewPhi}, {Pi, 0}}
    ];
    allfreqs = Union@Flatten@Table[{-(freqs - DetuningA[i] /. params), -(freqs - DetuningB[i] /. params)}, {i, nb}];
    vg = OptionValue@InitVG;
    If[vg === Automatic,
        vg = VelocityGroups1 @@ (
            {DopplerWidth, allfreqs, Sequence @@ (OptionValue[InitVGParams]{{NaturalWidth[2], 1}, {DopplerWidth, 1}}),
                ShowPlot -> OptionValue@PlotVelocityGroups} /. params
        )
    ];
    solve = buildAndSolve[sys, params, vg,
        {FluxReport -> Append[fluxReport, VelocitySpectrum], SkipLinearSolve -> OptionValue@SkipLinearSolve}
    ];
    speclist = VelocitySpectrum /. solve[[-1]] /. defaultViewDirection;
    If[!MemberQ[fluxReport, VelocitySpectrum], solve = DeleteCases[solve, VelocitySpectrum -> _, {2}]];
    resultTable = {solve};
    Table[
        spec = Interpolation[speclist, InterpolationOrder -> 1];
        bendfunc = Interpolation[bendList[speclist], InterpolationOrder -> 1];
        wf = nWeightFunction[spec, bendfunc, {5, 5, 5}, {x, -3dw, 3dw}];
        vg = VelocityGroups2[wf, {x, -3dw, 3dw}, nog[[i]], dw, 5 RecoilShift /. params, allfreqs,
            ShowPlot -> OptionValue@PlotVelocityGroups
        ];
        solve = buildAndSolve[sys, params, vg, FluxReport -> Append[fluxReport, VelocitySpectrum]];
        speclist = VelocitySpectrum /. solve[[-1]] /. defaultViewDirection;
        If[!MemberQ[fluxReport, VelocitySpectrum],
            solve = DeleteCases[solve, VelocitySpectrum -> _, {2}]
        ];
        AppendTo[resultTable, solve],
        {i, Length[nog]}
    ];
    If[Length@fluxReport == 0,
        If[OptionValue@ReturnAllSteps, resultTable[[All, 1]], resultTable[[-1, 1]]],
        If[OptionValue@ReturnAllSteps, resultTable, Last@resultTable]
    ]
];


(*ReturnFlux[syst_,par_,vg_,OptionsPattern[]]:=Module[{params,atdata,solve},
params=massageParams[par,Last[syst]];
solve=BuildAndSolve[syst,params,vg,FluxReport->Append[OptionValue[FluxReport],VelocitySpectrum]];
If[!MemberQ[OptionValue[FluxReport],VelocitySpectrum],solve=DeleteCases[solve,VelocitySpectrum->_,{2}]];If[MemberQ[OptionValue[FluxReport],VelocityGroups],solve=Insert[solve,VelocityGroups->vg,{-1,-1}]];
If[Length[OptionValue[FluxReport]]===0,solve[[1]],solve]
]*)


Options[buildAndSolve] = {FluxReport -> {}, SkipLinearSolve -> False};


buildAndSolve[{nvars_, A110_, A11S_, A11G_, A11R_, A12G_, A21R_, A12B_, b1G_, fluorvec1_, atdata_},
    params_, vg_, OptionsPattern[]] := Module[
    {nd, nb, fluorvecn, A110list, A110nList, A11Sn, A11Gn, A11Rn, A12Gn, A21Rn, A12Bn, bGn, A21RList,
        A12BList, blocks, fullAmat, diagprecmat, prec, fullbvec, fullfluorvec, sol, fl, vccfunc, vccrate1, vccrate2,
        vcr1, vcr2},
    (*Number of dithering regions*)
    nb = NDither /. params;
    (*Number of velocity groups*)
    nd = Length[vg];
    (*Function describing transfer rate between velocity groups due to collisions*)
    vccfunc = VccRateFunction /. params;
    (*Substitute numerical values into input matrices*)
    A110list = If[nb == 1,
        {A110},
        Table[
            SparseReplace[A110,
                {RabiFrequencyA[1] -> RabiFrequencyA[i], RabiFrequencyB[1] -> RabiFrequencyB[i],
                    DitherRegionFraction[1] -> DitherRegionFraction[i], DetuningA[1] -> DetuningA[i],
                    DetuningB[1] -> DetuningB[i], LightPhase[1] -> LightPhase[i]}
            ],
            {i, nb}
        ]
    ];
    {fluorvecn, A110nList, A11Sn, A11Gn, A11Rn, A12Gn, A21Rn, A12Bn, bGn} = N[
        SparseReplace[SparseReplace[{fluorvec1, A110list, A11S, A11G, A11R, A12G, A21R, A12B, b1G}, params], params]
    ];
    {groundpos, excitedpos} = Replace[{b1G, fluorvec1 /. {ViewTheta -> \[Pi], ViewPhi -> 0}}, x_ /; (x=!=0) -> 1, {2}];
    (*Check if matrices are numerical*)
    If[!And @@ (ArrayQ[#, _, NumericQ]& /@ {(*fluorvecn,*) A110nList, A11Sn, A11Gn, A11Rn, A12Gn, A21Rn, A12Bn, bGn}),
        Print["Nonnumerical matrix."];
        Abort[]
    ];
    (*A12List=Table[A12Gn vg[[i,2]],{i,nd}];*)
    A21RList = Table[A21Rn / (vg[[i, 1]] - vg[[i - 1, 1]]), {i, 2, nd}];
    A12BList = A12Bn #& /@ (Table[DitherRegionFraction[i] , {i, nb}] /. params);
    If[!And @@ (ArrayQ[#, _, NumericQ]& /@ {A21RList, A12BList}),
        Print["BuildAndSolve[]: Nonnumerical matrix (1), check the option list."];
        Abort[]
    ];
    (* For hard collisions (VCCRateFunction -> Automatic), the formulas for transfer rates can be simplified. We use
       this here to optimize the hard collision case. *)
    If[vccfunc === Automatic,
        vcr1 = (A11Gn(1 - #)& /@ vg[[All, 2]]) / (VccTime /. params);
        vcr2 = (A12Gn #& /@ vg[[All, 2]]) / (VccTime /. params);
        If[!And @@ (ArrayQ[#, _, NumericQ]& /@ {vcr1, vcr2}),
            Print["BuildAndSolve[]: Nonnumerical matrix (2), check the option list."];
            Abort[]
        ];
        vccrate1[i_] := vcr1[[i]];
        vccrate2[i_, j_] := vcr2[[i]],
        vccrate1[i_] := A11Gn(vccfunc[vg[[i, 1]], Infinity] - vccfunc[vg[[i, 1]], vg[[i, 3, 2]]] + vccfunc[vg[[i, 1]], vg[[i, 3, 1]]]);
        vccrate2[i_, j_] := A12Gn(1 / vg[[j, 2]] (vccfunc[vg[[i, 1]], vg[[j, 3, 2]]] - vccfunc[vg[[i, 1]], vg[[j, 3, 1]]])vg[[i, 2]])
    ];
    blocks = Table[
        Which[
            k == l,
            A110nList[[k]] + vg[[i, 1]] A11Sn + vccrate1[i] + If[i < nd, A11Rn / (vg[[i + 1, 1]] - vg[[i, 1]]), 0],
            k != l,
            A12BList[[k]]
        ],
        {i, nd}, {k, nb}, {l, nb}];
    fullAmat = ArrayFlatten@ArrayFlatten@Table[
        If[i == j,
            blocks[[i]],
            Table[
                If[
                    k == l,
                    vccrate2[i, j] + If[j == i - 1, A21RList[[j]], 0],
                    0.
                ],
                {k, nb}, {l, nb}]],
        {i, nd}, {j, nd}];
    diagprecmat = ArrayFlatten@ArrayFlatten@Table[
        If[i == j,
            blocks[[i]],
            0.],
        {i, nd}, {j, nd}];
    (* Change R.H. 02Sep09 *)
    If[OptionValue[SkipLinearSolve] == False,
        prec[x_] = LinearSolve[diagprecmat][x]
    ];
    fullbvec = Flatten@Table[Table[vg[[i, 2]]bGn, {nb}], {i, nd}];
    fullfluorvec = Flatten@Table[fluorvecn, {nd nb}];
    fullgroundpos = Flatten@Table[groundpos, {nd nb}];
    fullexcitedpos = Flatten@Table[excitedpos, {nd nb}];
    If[!And @@ (ArrayQ[#, _, NumericQ]& /@ {fullAmat, diagprecmat, fullbvec(*, fullfluorvec*)}),
        Print["BuildAndSolve[]: Nonnumerical matrix (3), check the option list."];
        Abort[]
    ];
    sol = If[OptionValue[SkipLinearSolve] == False,
        LinearSolve[fullAmat, fullbvec, Method -> {Krylov, Preconditioner -> prec}],
        0
    ];
    fl = Re[fullfluorvec.sol];
    {fl, OptionValue@FluxReport /.
        {VelocitySpectrum -> (VelocitySpectrum ->
            MapThread[
                    {#3, #1 / (-Subtract @@ #2)}&,
                    {Total /@ Partition[Total /@ Partition[fullfluorvec sol, nvars], nb], vg[[All, 3]], vg[[All, 1]]}
                ]
            ),
            VelocityGroups -> (VelocityGroups -> vg),
            DMSolution -> (DMSolution -> sol),
            BlochMatrix -> (BlochMatrix -> {fullAmat, fullbvec, fullfluorvec}),
            PopulationPositions -> (PopulationPositions -> {fullgroundpos, fullexcitedpos}),
            RFParams -> (RFParams -> params)
        }
    }
];


buildLightDependentPart[nvars_, A110_, atdata_, par_, vg_] := Module[
    {nd, nb, A110list, A110nList, blocks, params, params0, params1},
    params = Join[par, DefaultParameters /. Options[PsiMeso]];
    params0 =
        If[(RepumpFraction /. params) > 0,
            Join[PsiMesoOptDefault /. params, params],
            Join[PsiMesoNoRepumpOptDefault /. params, params],
            Print["BuildLightDependentPart[]: Cannot decide if RepumpFraction is nonzero"];
            Abort[]
        ];
    params1 = massageParams[params0, atdata];
    (*Number of dithering regions*)
    nb = NDither /. params1;
    (*Number of velocity groups*)
    nd = Length[vg];
    (*Substitute numerical values into input matrices*)
    A110list = If[nb == 1,
        {A110},
        Table[
            SparseReplace[A110, {
                RabiFrequencyA[1] -> RabiFrequencyA[i], RabiFrequencyB[1] -> RabiFrequencyB[i],
                DetuningA[1] -> DetuningA[i], DetuningB[1] -> DetuningB[i], LightPhase[1] -> LightPhase[i]}
            ],
            {i, nb}
        ]
    ];
    {A110nList} = N[SparseReplace[SparseReplace[{A110list}, params1], params1]];
    (*Check if matrices are numerical*)
    If[!And @@ (ArrayQ[#, _, NumericQ]& /@ {A110nList}),
        Print["Nonnumerical matrix."];
        Abort[]
    ];
    blocks = Table[
        If[
            k == l,
            A110nList[[k]],
            0.
        ],
        {i, nd}, {k, nb}, {l, nb}];
    ArrayFlatten@ArrayFlatten@Table[
        If[i == j,
            blocks[[i]],
            0.],
        {i, nd},
        {j, nd}
    ]
];


(* ::Subsection:: *)
(*LGS Constants and Default Parameters*)


BGParanal = 0.22750;  (* geomagnetic field in Paranal at altitude 90km in 2009 [G] from [MSIS90]*)
BGSOR = 0.48052;  (* geomagnetic field at SOR at altitude 90km in 2009 [G] from [MSIS90]*)
BGWendelstein = 0.46044;  (* geomagnetic field over Wendelstein at altutude 90km in 2009 [G] from [MSIS90]*)
BGMaunaKea = 0.35;  (* geomagnetic field over Mauna Kea at altitude 90km in 2009 [G] from [MSIS90]*)
\[CapitalOmega]NaRecoil = 2\[Pi] 50. 10^3;  (* recoil frequency of sodium atoms at 589nm *)
(*\[Gamma]2FWHM=Sqrt[-1.0+Sqrt[2]];  * Conversion factor from Lorentzian-square \[Gamma] to FWHM *)
fOffsetD2a = -0.6268 * 10^9; (* Offset of D2a line from D2 line center *)
fOffsetD2ab = 1.710 * 10^9;(* Optimal frequency offset between D2a and D2b, modified on 17Nov11 with the introduction of DopplerFlip *)


(* ::Text:: *)
(*Laser Parameters:*)
(*	IMeso: Mesospheric laser intensity [W/m^2]*)
(*	RepumpFraction: Repumping intensity fraction*)
(*	LightPolarization: Normalized ellipticity angle (LightPolarization := \[Epsilon]/(\[Pi]/4))*)
(*	LightWavelength: Laser center wavelength [m]*)
(*	Phase: Overall light phase.*)
(*	FWHMbw: Laser FWHM bandwidth [Hz]*)
(*	\[CapitalDelta]fLaser: Laser line detuning from optimum (both D2a and D2b) [Hz]*)
(*	\[CapitalDelta]fLaserDabOff: Offset of D2a--D2b frequency spacing from fOffsetD2ab [Hz]*)


laserOptsDefault = {IMeso -> 46.0, RepumpFraction -> 0.12, LightPolarization -> 1.0, LightWavelength -> 589.159 * 10^-9,
    Phase -> 0.0, FWHMbw -> 0.0, \[CapitalDelta]fLaser -> 0.0, \[CapitalDelta]fLaserDabOff -> 0.0, LaserWidth -> 0};


(* ::Text:: *)
(*Atmospheric and Mesospheric Parameters:*)
(*	BG: Mesospheric geomagnetic field strength [G]*)
(*	Ta: One-way transmission at \[Lambda] at zenith*)
(*	Temperature: Mesospheric temperature [K]*)
(*	tauCollMult: Multiplier for S- and v-damping collision times*)
(*	HNa: Altitude of sodium centroid above sea level [m]*)
(*	vrms\[Gamma]: rms velocity of diffusion effects [m/s], \[Gamma] = vrms\[Gamma] / (DLT/\[Mu]LT)*)


atmoOptsDefault = {BG -> BGParanal, Ta -> 0.84, Temperature -> 185.0, SDampingTime -> 245.0 * 10^-6,
    VccTime -> 35.0 * 10^-6, HNa -> 92.0 * 10^3, vrms\[Gamma] -> 38.0, RecoilShift -> \[CapitalOmega]NaRecoil, VccRateFunction -> Automatic};


(* ::Text:: *)
(*Launch Telescope System (LTS) Parameters:*)
(*	PLaunch: Launched power [W]*)
(*	Zenith: Zenith angle [rad]*)
(*	Htele: Altitude of launch telescope above sea level [m]*)
(*	DLT: Launch telescope aperture [m]*)
(*	\[Mu]LT: Ratio of DLT to 1/e^2 lauch beam intensity diameter (not radius!), see Yura, Eq.(2.7)*)
(*	MsqrLT: M^2 beam quality factor out of the launch telescope*)
(*	MagneticZenith, MagneticAzimuth: elevation and azimuth angles of B vector in a system where the laser points along z *)


LTSOptsDefault = {PLaunch -> 20.0, Zenith -> 30.0 * \[Pi] / 180.0, Htele -> 2650.0, DLT -> 0.4, \[Mu]LT -> 50.0 / 36.0,
    MsqrLT -> 1.1, MagneticZenith -> \[Pi] / 2.0, MagneticAzimuth -> \[Pi] / 2.0, ViewTheta -> Pi, ViewPhi -> 0};


(*  WARNING: NaLGSDefaults and PhiOptDefault depend on the three preceding option sets! *)


NaLGSDefaults = Join[laserOptsDefault, atmoOptsDefault, LTSOptsDefault];


$DefaultLGSParameters = NaLGSDefaults;


PsiMesoOptDefault = {NDither -> 1 + Sign[FWHMbw], DitherRate -> Sign[FWHMbw] 2 * \[Pi] * FWHMbw, DitherDetuning[1] -> 0,
    LightPhase[1] -> Phase, LightPhase[2] -> Phase + \[Pi], DitherCoherenceTransfer -> 0,
    EllipticityA -> LightPolarization * \[Pi] / 4, DetuningA -> 2 \[Pi] (\[CapitalDelta]fLaser + fOffsetD2a),
    LightIntensityA -> (1.0 - RepumpFraction)IMeso, EllipticityB -> LightPolarization * \[Pi] / 4,
    DetuningB -> 2 \[Pi] (\[CapitalDelta]fLaser + fOffsetD2a + fOffsetD2ab + \[CapitalDelta]fLaserDabOff), LightIntensityB -> RepumpFraction IMeso,
    AtomicDensity -> 1, BeamTransitRate -> vrms\[Gamma] / (DLT / \[Mu]LT), SDampingRate -> 1 / SDampingTime,
    LarmorFrequency -> LandeGF BohrMagneton BG Gauss};


(* ::Text:: *)
(*This one is for no repumping. In this case, we insert the power also into the D2b line.*)


PsiMesoNoRepumpOptDefault = {NDither -> 1 + Sign[FWHMbw], DitherRate -> Sign[FWHMbw]2 * \[Pi] * FWHMbw, DitherDetuning[1] -> 0,
    LightPhase[1] -> Phase, LightPhase[2] -> Phase + \[Pi], DitherCoherenceTransfer -> 0,
    EllipticityA -> LightPolarization * \[Pi] / 4, DetuningA -> 2 \[Pi] (\[CapitalDelta]fLaser + fOffsetD2a), LightIntensityA -> IMeso,
    EllipticityB -> LightPolarization * \[Pi] / 4, DetuningB -> 2 \[Pi](\[CapitalDelta]fLaser + fOffsetD2a), LightIntensityB -> IMeso,
    AtomicDensity -> 1, BeamTransitRate -> vrms\[Gamma] / (DLT / \[Mu]LT), SDampingRate -> 1 / SDampingTime,
    LarmorFrequency -> LandeGF BohrMagneton BG Gauss};


(* ::Subsection:: *)
(*PsiMeso*)


Options[PsiMeso] = Join[{DefaultParameters :> $DefaultLGSParameters}, Options@ReturnFlux];
SetOptions[PsiMeso, RefineResult -> {100}];


PsiMeso[syst_, params_, opts : OptionsPattern[]] := Module[{flux, imes, params1, rfopts},
    params1 = Join[params, OptionValue@DefaultParameters];
    rfopts = FilterRules[{opts, Options@PsiMeso}, Options@ReturnFlux];
    params1 = If[Positive@(RepumpFraction /. params1),
        Join[PsiMesoOptDefault /. params1, params1],
        Join[PsiMesoNoRepumpOptDefault /. params1, params1],
        Print["PsiMeso[]: Cannot decide if RepumpFraction is nonzero, RepumpFraction: ", RepumpFraction /. params1];
        Abort[]
    ];
    flux = If[
        Positive[imes = IMeso /. params1],
        ReturnFlux[syst, params1, rfopts],
        0.0,
        Print["PsiMeso[]: Cannot decide if IMeso is nonzero, IMeso: ", IMeso /. params1];
        Abort[]
    ];
    If[NumericQ@flux, flux / imes, flux / {imes, 1}]
];

End[];

EndPackage[];

